from datetime import datetime, timedelta

import pytest
from croniters import croniter


@pytest.fixture
def epoch():
    return datetime(2020, 1, 1, 0, 0)


def test_random(epoch):
    """Test random definition"""
    obj = croniter('R R * * *', epoch)
    result_1 = obj.get_next(datetime)
    assert result_1 >= datetime(2020, 1, 1, 0, 0)
    assert result_1 <= datetime(2020, 1, 1, 0, 0) + timedelta(days=1)
    result_2 = obj.get_next(datetime)
    assert result_2 >= datetime(2020, 1, 2, 0, 0)
    assert result_2 <= datetime(2020, 1, 2, 0, 0) + timedelta(days=1)


def test_random_range(epoch):
    """Test random definition within a range"""
    obj = croniter('R R R(10-20) * *', epoch)
    result_1 = obj.get_next(datetime)
    assert result_1 >= datetime(2020, 1, 10, 0, 0)
    assert result_1 <= datetime(2020, 1, 10, 0, 0) + timedelta(days=11)
    result_2 = obj.get_next(datetime)
    assert result_2 >= datetime(2020, 2, 10, 0, 0)
    assert result_2 <= datetime(2020, 2, 10, 0, 0) + timedelta(days=11)


def test_random_float(epoch: datetime):
    """Test random definition, float result"""
    obj = croniter('R R * * *', epoch)
    result_1 = obj.get_next(float)
    assert isinstance(result_1, float)
    assert result_1 >= 1577836800.0
    assert result_1 <= 1577836800.0 + (60 * 60 * 24)
    result_2 = obj.get_next(float)
    assert isinstance(result_2, float)
    assert result_2 >= 1577923200.0
    assert result_2 <= 1577923200.0 + (60 * 60 * 24)


def test_random_with_year(epoch):
    """Test random with year range"""
    obj = croniter('* * * * * * R(2025-2030)', epoch)
    result = obj.get_next(datetime)
    assert result.year >= 2025
    assert result.year <= 2030



import os
import time
from collections.abc import Generator
from datetime import datetime

import pytest
from croniters import croniter


@pytest.fixture
def timezone_setup() -> Generator[tuple[datetime, list[datetime]], None, None]:
    orig_time = os.environ.setdefault('TZ', '')
    base = datetime(2024, 1, 25, 4, 46)
    expected = [
        datetime(2024, 1, 25, 4, 50),
        datetime(2024, 1, 25, 4, 55),
        datetime(2024, 1, 25, 5, 0),
    ]

    yield base, expected

    if not orig_time:
        del os.environ['TZ']
    else:
        os.environ['TZ'] = orig_time
    time.tzset()


@pytest.mark.parametrize('tz', ['UTC', 'UTC-8'])
def test_dst_138(timezone_setup, tz):
    """Test DST handling - see issue #138"""
    base, expected = timezone_setup

    os.environ['TZ'] = tz
    time.tzset()

    itr = croniter('*/5 * * * *', base)
    result = [itr.get_next(datetime) for _ in range(3)]
    assert result == expected



from datetime import datetime

import pytest
import pytz
from croniters import (
    CroniterBadCronError,
    CroniterBadDateError,
    CroniterBadTypeRangeError,
    croniter,
    croniter_range,
)


class mydatetime(datetime):
    """Custom datetime class for testing"""

    pass


def test_1day_step():
    start = datetime(2016, 12, 2)
    stop = datetime(2016, 12, 10)
    fwd = list(croniter_range(start, stop, '0 0 * * *'))
    assert len(fwd) == 9
    assert fwd[0] == start
    assert fwd[-1] == stop
    # Test the same, but in reverse
    rev = list(croniter_range(stop, start, '0 0 * * *'))
    assert len(rev) == 9
    # Ensure forward/reverse are a mirror image
    rev.reverse()
    assert fwd == rev


def test_1day_step_no_ends():
    # Test without ends (exclusive)
    start = datetime(2016, 12, 2)
    stop = datetime(2016, 12, 10)
    fwd = list(croniter_range(start, stop, '0 0 * * *', exclude_ends=True))
    assert len(fwd) == 7
    assert fwd[0] != start
    assert fwd[-1] != stop
    # Test the same, but in reverse
    rev = list(croniter_range(stop, start, '0 0 * * *', exclude_ends=True))
    assert len(rev) == 7
    assert fwd[0] != stop
    assert fwd[-1] != start


def test_1month_step():
    start = datetime(1982, 1, 1)
    stop = datetime(1983, 12, 31)
    res = list(croniter_range(start, stop, '0 0 1 * *'))
    assert len(res) == 24
    assert res[0] == start
    assert res[5].day == 1
    assert res[-1] == datetime(1983, 12, 1)


def test_1minute_step_float():
    start = datetime(2000, 1, 1, 0, 0)
    stop = datetime(2000, 1, 1, 0, 1)
    res = list(croniter_range(start, stop, '* * * * *', ret_type=float))
    assert len(res) == 2
    assert res[0] == 946684800.0
    assert res[-1] - res[0] == 60


def test_auto_ret_type():
    data = [
        (datetime(2019, 1, 1), datetime(2020, 1, 1), datetime),
        (1552252218.0, 1591823311.0, float),
    ]
    for start, stop, rtype in data:
        ret = list(croniter_range(start, stop, '0 0 * * *'))
        assert isinstance(ret[0], rtype)


def test_input_type_exceptions():
    dt_start1 = datetime(2019, 1, 1)
    dt_stop1 = datetime(2020, 1, 1)
    f_start1 = 1552252218.0
    f_stop1 = 1591823311.0
    # Mix start/stop types
    with pytest.raises(TypeError):
        list(croniter_range(dt_start1, f_stop1, '0 * * * *'), ret_type=datetime)
    with pytest.raises(TypeError):
        list(croniter_range(f_start1, dt_stop1, '0 * * * *'))


def test_timezone_dst():
    """Test across DST transition, which technically is a timezone change."""
    tz = pytz.timezone('US/Eastern')
    start = tz.localize(datetime(2020, 10, 30))
    stop = tz.localize(datetime(2020, 11, 10))
    res = list(croniter_range(start, stop, '0 0 * * *'))
    assert res[0].tzinfo != res[-1].tzinfo
    assert len(res) == 12


def test_extra_hour_day_prio():
    def datetime_tz(*args, **kw):
        """Defined this in another branch. single-use-version"""
        tzinfo = kw.pop('tzinfo')
        return tzinfo.localize(datetime(*args))

    tz = pytz.timezone('US/Eastern')
    cron = '0 3 * * *'
    start = datetime_tz(2020, 3, 7, tzinfo=tz)
    end = datetime_tz(2020, 3, 11, tzinfo=tz)
    ret = [i.isoformat() for i in croniter_range(start, end, cron)]
    assert ret == [
        '2020-03-07T03:00:00-05:00',
        '2020-03-08T03:00:00-04:00',
        '2020-03-09T03:00:00-04:00',
        '2020-03-10T03:00:00-04:00',
    ]


def test_issue145_getnext():
    # Example of quarterly event cron schedule
    start = datetime(2020, 9, 24)
    cron = '0 13 8 1,4,7,10 wed'
    with pytest.raises(CroniterBadDateError):
        it = croniter(cron, start, day_or=False, max_years_between_matches=1)
        it.get_next()
    # New functionality (0.3.35) allowing croniter to find sparse matches of cron patterns across multiple years
    it = croniter(cron, start, day_or=False, max_years_between_matches=5)
    assert it.get_next(datetime) == datetime(2025, 1, 8, 13)


def test_issue145_range():
    cron = '0 13 8 1,4,7,10 wed'
    matches = list(
        croniter_range(datetime(2020, 1, 1), datetime(2020, 12, 31), cron, day_or=False)
    )
    assert len(matches) == 3
    assert matches[0] == datetime(2020, 1, 8, 13)
    assert matches[1] == datetime(2020, 4, 8, 13)
    assert matches[2] == datetime(2020, 7, 8, 13)

    # No matches within this range; therefore expect empty list
    matches = list(
        croniter_range(
            datetime(2020, 9, 30), datetime(2020, 10, 30), cron, day_or=False
        )
    )
    assert len(matches) == 0


def test_croniter_range_derived_class():
    # trivial example extending croniter
    class croniter_nosec(croniter):
        """Like croniter, but it forbids second-level cron expressions."""

        @classmethod
        def expand(cls, expr_format, *args, **kwargs):
            if len(expr_format.split()) == 6:
                raise CroniterBadCronError("Expected 'min hour day mon dow'")
            return croniter.expand(expr_format, *args, **kwargs)

    cron = '0 13 8 1,4,7,10 wed'
    matches = list(
        croniter_range(
            datetime(2020, 1, 1),
            datetime(2020, 12, 31),
            cron,
            day_or=False,
            _croniter=croniter_nosec,
        )
    )
    assert len(matches) == 3

    cron = '0 1 8 1,15,L wed 15,45'
    with pytest.raises(CroniterBadCronError):
        # Should fail using the custom class that forbids the seconds expression
        croniter_nosec(cron)

    with pytest.raises(CroniterBadCronError):
        # Should similarly fail because the custom class rejects seconds expr
        i = croniter_range(
            datetime(2020, 1, 1),
            datetime(2020, 12, 31),
            cron,
            _croniter=croniter_nosec,
        )
        next(i)


def test_dt_types():
    start = mydatetime(2020, 9, 24)
    stop = datetime(2020, 9, 28)
    try:
        list(croniter_range(start, stop, '0 0 * * *'))
    except CroniterBadTypeRangeError:
        pytest.fail('should not be triggered')


def test_configure_second_location():
    start = datetime(2016, 12, 2, 0, 0, 0)
    stop = datetime(2016, 12, 2, 0, 1, 0)
    fwd = list(croniter_range(start, stop, '*/20 * * * * *', second_at_beginning=True))
    assert len(fwd) == 4
    assert fwd[0] == start
    assert fwd[-1] == stop


def test_year_range():
    start = datetime(2010, 1, 1)
    stop = datetime(2030, 1, 1)
    fwd = list(croniter_range(start, stop, '0 0 1 1 ? 0 2020-2024,2028'))
    assert len(fwd) == 6
    assert fwd[0] == datetime(2020, 1, 1)
    assert fwd[-1] == datetime(2028, 1, 1)



import os
import sys
from datetime import datetime
from timeit import Timer

import pytest
import pytz
from croniters import croniter


def test_run_long(iterations=1):  # noqa: C901
    dt = datetime(2010, 1, 23, 12, 18)
    itr = croniter('*/1 * * * *', dt)
    for _ in range(iterations):  # ~ 58
        itr.get_next()

    itr = croniter('*/5 * * * *', dt)
    for _ in range(iterations):
        itr.get_next()

    dt = datetime(2010, 1, 24, 12, 2)
    itr = croniter('0 */3 * * *', dt)
    for _ in range(iterations):
        itr.get_next()

    dt = datetime(2010, 2, 24, 12, 9)
    itr = croniter('0 0 */3 * *', dt)
    for _ in range(iterations):
        itr.get_next(datetime)

    # test leap year
    dt = datetime(1996, 2, 27)
    itr = croniter('0 0 * * *', dt)
    for _ in range(iterations):
        itr.get_next(datetime)

    dt2 = datetime(2000, 2, 27)
    itr2 = croniter('0 0 * * *', dt2)
    for _ in range(iterations):
        itr2.get_next(datetime)

    dt = datetime(2010, 2, 25)
    itr = croniter('0 0 * * sat', dt)
    for _ in range(iterations):
        itr.get_next(datetime)

    dt = datetime(2010, 1, 25)
    itr = croniter('0 0 1 * wed', dt)
    for _ in range(iterations):
        itr.get_next(datetime)

    dt = datetime(2010, 1, 25)
    itr = croniter('0 0 1 * *', dt)
    for _ in range(iterations):
        itr.get_next()

    dt = datetime(2010, 8, 25, 15, 56)
    itr = croniter('*/1 * * * *', dt)
    for _ in range(iterations):
        itr.get_prev(datetime)

    dt = datetime(2010, 8, 25, 15, 0)
    itr = croniter('*/1 * * * *', dt)
    for _ in range(iterations):
        itr.get_prev(datetime)

    dt = datetime(2010, 8, 25, 0, 0)
    itr = croniter('*/1 * * * *', dt)
    for _ in range(iterations):
        itr.get_prev(datetime)

    dt = datetime(2010, 8, 25, 15, 56)
    itr = croniter('0 0 * * sat,sun', dt)
    for _ in range(iterations):
        itr.get_prev(datetime)

    dt = datetime(2010, 2, 25)
    itr = croniter('0 0 * * 7', dt)
    for _ in range(iterations):
        itr.get_prev(datetime)

    # dst regression test
    tz = pytz.timezone('Europe/Bucharest')
    offsets = set()
    dst_cron = '15 0,3 * 3 *'
    dst_iters = int(2 * 31 * (iterations / 40))
    dt = datetime(2010, 1, 25, tzinfo=tz)
    itr = croniter(dst_cron, dt)
    for _ in range(dst_iters):
        d = itr.get_next(datetime)
        offsets.add(d.utcoffset())
    itr = croniter(dst_cron, dt)
    for _ in range(dst_iters):
        d = itr.get_prev(datetime)
        offsets.add(d.utcoffset())


@pytest.mark.skipif(sys.version_info[0] < 3, reason='Python 2 not supported')
def test_not_long_time():
    iterations = int(os.environ.get('CRONITER_TEST_SPEED_ITERATIONS', '40'))

    # Create namespace for Timer
    namespace = {'run_long_test': test_run_long, 'iterations': iterations}

    t = Timer('run_long_test(iterations)', globals=namespace)
    limit = 80
    ret = t.timeit(limit)
    assert ret < limit, f'Regression in croniter speed detected ({ret} {limit}).'



"""Type stubs for croniters Rust extension."""

MINUTE_FIELD: int
HOUR_FIELD: int
DAY_FIELD: int
MONTH_FIELD: int
DOW_FIELD: int
SECOND_FIELD: int
YEAR_FIELD: int

M_ALPHAS: dict[str, int]
DOW_ALPHAS: dict[str, int]
UNIX_FIELDS: list[int]
SECOND_FIELDS: list[int]
YEAR_FIELDS: list[int]



import random
import uuid
from datetime import datetime, timedelta

import pytest
from croniters import CroniterBadCronError, CroniterNotAlphaError, croniter


# Convert base class setup into fixtures
@pytest.fixture
def hash_setup():
    return {'epoch': datetime(2020, 1, 1, 0, 0), 'hash_id': 'hello'}


# Helper function to replace the _test_iter method
def check_iter(
    definition, expectations, delta, epoch=None, hash_id=None, next_type=None
):
    if epoch is None:
        epoch = datetime(2020, 1, 1, 0, 0)
    if hash_id is None:
        hash_id = 'hello'
    if next_type is None:
        next_type = datetime
    if not isinstance(expectations, (list, tuple)):
        expectations = (expectations,)

    obj = croniter(definition, epoch, hash_id=hash_id)
    testval = obj.get_next(next_type)
    assert testval in expectations
    if delta is not None:
        assert obj.get_next(next_type) == testval + delta


# Convert test methods to functions
def test_hash_hourly(hash_setup):
    """Test manually-defined hourly"""
    check_iter('H * * * *', datetime(2020, 1, 1, 0, 10), timedelta(hours=1))


def test_hash_daily(hash_setup):
    """Test manually-defined daily"""
    check_iter('H H * * *', datetime(2020, 1, 1, 11, 10), timedelta(days=1))


def test_hash_weekly(hash_setup):
    """Test manually-defined weekly"""
    # croniter 1.0.5 changes the defined weekly range from (0, 6)
    # to (0, 7), to match cron's behavior that Sunday is 0 or 7.
    # This changes the hash, so test for either.
    check_iter(
        'H H * * H',
        (datetime(2020, 1, 3, 11, 10), datetime(2020, 1, 5, 11, 10)),
        timedelta(weeks=1),
    )


def test_hash_monthly(hash_setup):
    """Test manually-defined monthly"""
    check_iter('H H H * *', datetime(2020, 1, 1, 11, 10), timedelta(days=31))


def test_hash_yearly(hash_setup):
    """Test manually-defined yearly"""
    check_iter('H H H H *', datetime(2020, 9, 1, 11, 10), timedelta(days=365))


def test_hash_second(hash_setup):
    """Test seconds

    If a sixth field is provided, seconds are included in the datetime()
    """
    check_iter('H H * * * H', datetime(2020, 1, 1, 11, 10, 32), timedelta(days=1))


def test_hash_year(hash_setup):
    """Test years

    provide a seventh field as year
    """
    check_iter('H H * * * H H', datetime(2066, 1, 1, 11, 10, 32), timedelta(days=1))


def test_hash_id_change(hash_setup):
    """Test a different hash_id returns different results given same definition and epoch"""
    check_iter('H H * * *', datetime(2020, 1, 1, 11, 10), timedelta(days=1))
    check_iter(
        'H H * * *',
        datetime(2020, 1, 1, 0, 24),
        timedelta(days=1),
        hash_id='different id',
    )


def test_hash_epoch_change(hash_setup):
    """Test a different epoch returns different results given same definition and hash_id"""
    check_iter('H H * * *', datetime(2020, 1, 1, 11, 10), timedelta(days=1))
    check_iter(
        'H H * * *',
        datetime(2011, 11, 12, 11, 10),
        timedelta(days=1),
        epoch=datetime(2011, 11, 11, 11, 11),
    )


def test_hash_range(hash_setup):
    """Test a hashed range definition"""
    check_iter('H H H(3-5) * *', datetime(2020, 1, 5, 11, 10), timedelta(days=31))
    check_iter(
        'H H * * * 0 H(2025-2030)', datetime(2029, 1, 1, 11, 10), timedelta(days=1)
    )


def test_hash_division(hash_setup):
    """Test a hashed division definition"""
    check_iter('H H/3 * * *', datetime(2020, 1, 1, 2, 10), timedelta(hours=3))
    check_iter(
        'H H H H * H H/2', datetime(2020, 9, 1, 11, 10, 32), timedelta(days=365 * 2)
    )


def test_hash_range_division(hash_setup):
    """Test a hashed range + division definition"""
    check_iter(
        'H(30-59)/10 H * * *', datetime(2020, 1, 1, 11, 30), timedelta(minutes=10)
    )


def test_hash_invalid_range(hash_setup):
    """Test validation logic for range_begin and range_end values"""
    try:
        check_iter(
            'H(11-10) H * * *', datetime(2020, 1, 1, 11, 31), timedelta(minutes=10)
        )
    except CroniterBadCronError as ex:
        assert str(ex) == 'Range end must be greater than range begin'


def test_hash_id_bytes(hash_setup):
    """Test hash_id as a bytes object"""
    check_iter(
        'H H * * *',
        datetime(2020, 1, 1, 14, 53),
        timedelta(days=1),
        hash_id=b'\x01\x02\x03\x04',
    )


def test_hash_float(hash_setup):
    """Test result as a float object"""
    check_iter('H H * * *', 1577877000.0, (60 * 60 * 24), next_type=float)


def test_invalid_definition(hash_setup):
    """Test an invalid definition raises CroniterNotAlphaError"""
    with pytest.raises(CroniterNotAlphaError):
        croniter('X X * * *', hash_setup['epoch'], hash_id=hash_setup['hash_id'])


def test_invalid_hash_id_type(hash_setup):
    """Test an invalid hash_id type raises TypeError"""
    with pytest.raises(TypeError):
        croniter('H H * * *', hash_setup['epoch'], hash_id={1: 2})


def test_invalid_divisor(hash_setup):
    """Test an invalid divisor type raises CroniterBadCronError"""
    with pytest.raises(CroniterBadCronError):
        croniter('* * H/0 * *', hash_setup['epoch'], hash_id=hash_setup['hash_id'])


def test_hash_word_midnight(hash_setup):
    """Test built-in @midnight

    @midnight is actually up to 3 hours after midnight, not exactly midnight
    """
    check_iter('@midnight', datetime(2020, 1, 1, 2, 10, 32), timedelta(days=1))


def test_hash_word_hourly(hash_setup):
    """Test built-in @hourly"""
    check_iter('@hourly', datetime(2020, 1, 1, 0, 10, 32), timedelta(hours=1))


def test_hash_word_daily(hash_setup):
    """Test built-in @daily"""
    check_iter('@daily', datetime(2020, 1, 1, 11, 10, 32), timedelta(days=1))


def test_hash_word_weekly(hash_setup):
    """Test built-in @weekly"""
    # croniter 1.0.5 changes the defined weekly range from (0, 6)
    # to (0, 7), to match cron's behavior that Sunday is 0 or 7.
    # This changes the hash, so test for either.
    check_iter(
        '@weekly',
        (datetime(2020, 1, 3, 11, 10, 32), datetime(2020, 1, 5, 11, 10, 32)),
        timedelta(weeks=1),
    )


def test_hash_word_monthly(hash_setup):
    """Test built-in @monthly"""
    check_iter('@monthly', datetime(2020, 1, 1, 11, 10, 32), timedelta(days=31))


def test_hash_word_yearly(hash_setup):
    """Test built-in @yearly"""
    check_iter('@yearly', datetime(2020, 9, 1, 11, 10, 32), timedelta(days=365))


def test_hash_word_annually(hash_setup):
    """Test built-in @annually

    @annually is the same as @yearly
    """
    obj_annually = croniter(
        '@annually', hash_setup['epoch'], hash_id=hash_setup['hash_id']
    )
    obj_yearly = croniter('@yearly', hash_setup['epoch'], hash_id=hash_setup['hash_id'])
    assert obj_annually.get_next(datetime) == obj_yearly.get_next(datetime)
    assert obj_annually.get_next(datetime) == obj_yearly.get_next(datetime)


@pytest.fixture
def hash_ids():
    _rd = random.Random()
    _rd.seed(100)
    return [uuid.UUID(int=_rd.getrandbits(128)).bytes for _ in range(350)]


# Minutes expander tests
def test_expand_minutes(hash_ids):
    minutes = set()
    expression = 'H * * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        minutes.add(expanded[0][0][0])
    assert len(minutes) == 60
    assert min(minutes) == 0
    assert max(minutes) == 59


def test_expand_minutes_range_2_minutes(hash_ids):
    minutes = set()
    expression = 'H/2 * * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _minutes = expanded[0][0]
        assert len(_minutes) == 30
        minutes.update(_minutes)
    assert len(minutes) == 60
    assert min(minutes) == 0
    assert max(minutes) == 59


def test_expand_minutes_range_3_minutes(hash_ids):
    minutes = set()
    expression = 'H/3 * * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _minutes = expanded[0][0]
        assert len(_minutes) == 20
        minutes.update(_minutes)
    assert len(minutes) == 60
    assert min(minutes) == 0
    assert max(minutes) == 59


def test_expand_minutes_range_15_minutes(hash_ids):
    minutes = set()
    expression = 'H/15 * * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _minutes = expanded[0][0]
        assert len(_minutes) == 4
        minutes.update(_minutes)
    assert len(minutes) == 60
    assert min(minutes) == 0
    assert max(minutes) == 59


def test_expand_minutes_with_full_range(hash_ids):
    minutes = set()
    expression = 'H(0-59) * * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        minutes.add(expanded[0][0][0])
    assert len(minutes) == 60
    assert min(minutes) == 0
    assert max(minutes) == 59


# Hours expander tests
def test_expand_hours(hash_ids):
    hours = set()
    expression = 'H H * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        hours.add(expanded[0][1][0])
    assert len(hours) == 24
    assert min(hours) == 0
    assert max(hours) == 23


def test_expand_hours_range_every_2_hours(hash_ids):
    hours = set()
    expression = 'H H/2 * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _hours = expanded[0][1]
        assert len(_hours) == 12
        hours.update(_hours)
    assert len(hours) == 24
    assert min(hours) == 0
    assert max(hours) == 23


def test_expand_hours_range_4_hours(hash_ids):
    hours = set()
    expression = 'H H/4 * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _hours = expanded[0][1]
        assert len(_hours) == 6
        hours.update(_hours)
    assert len(hours) == 24
    assert min(hours) == 0
    assert max(hours) == 23


def test_expand_hours_range_8_hours(hash_ids):
    hours = set()
    expression = 'H H/8 * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _hours = expanded[0][1]
        assert len(_hours) == 3
        hours.update(_hours)
    assert len(hours) == 24
    assert min(hours) == 0
    assert max(hours) == 23


def test_expand_hours_range_10_hours(hash_ids):
    hours = set()
    expression = 'H H/10 * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _hours = expanded[0][1]
        assert len(_hours) in {2, 3}
        hours.update(_hours)
    assert len(hours) == 24
    assert min(hours) == 0
    assert max(hours) == 23


def test_expand_hours_range_12_hours(hash_ids):
    hours = set()
    expression = 'H H/12 * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _hours = expanded[0][1]
        assert len(_hours) == 2
        hours.update(_hours)
    assert len(hours) == 24
    assert min(hours) == 0
    assert max(hours) == 23


def test_expand_hours_with_full_range(hash_ids):
    minutes = set()
    expression = '* H(0-23) * * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        minutes.add(expanded[0][1][0])
    assert len(minutes) == 24
    assert min(minutes) == 0
    assert max(minutes) == 23


# Month days expander tests
def test_expand_month_days(hash_ids):
    month_days = set()
    expression = 'H H H * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        month_days.add(expanded[0][2][0])
    assert len(month_days) == 31
    assert min(month_days) == 1
    assert max(month_days) == 31


def test_expand_month_days_range_2_days(hash_ids):
    month_days = set()
    expression = '0 0 H/2 * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _days = expanded[0][2]
        assert len(_days) in {15, 16}
        month_days.update(_days)
    assert len(month_days) == 31
    assert min(month_days) == 1
    assert max(month_days) == 31


def test_expand_month_days_range_5_days(hash_ids):
    month_days = set()
    expression = 'H H H/5 * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _days = expanded[0][2]
        assert len(_days) in {6, 7}
        month_days.update(_days)
    assert len(month_days) == 31
    assert min(month_days) == 1
    assert max(month_days) == 31


def test_expand_month_days_range_12_days(hash_ids):
    month_days = set()
    expression = 'H H H/12 * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _days = expanded[0][2]
        assert len(_days) in {2, 3}
        month_days.update(_days)
    assert len(month_days) == 31
    assert min(month_days) == 1
    assert max(month_days) == 31


def test_expand_month_days_with_full_range(hash_ids):
    month_days = set()
    expression = '* * H(1-31) * *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        month_days.add(expanded[0][2][0])
    assert len(month_days) == 31
    assert min(month_days) == 1
    assert max(month_days) == 31


# Month expander tests
def test_expand_months(hash_ids):
    months = set()
    expression = 'H H * H *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        months.add(expanded[0][3][0])
    assert len(months) == 12
    assert min(months) == 1
    assert max(months) == 12


def test_expand_month_days_range_2_months(hash_ids):
    months = set()
    expression = 'H H * H/2 *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _months = expanded[0][3]
        assert len(_months) == 6
        months.update(_months)
    assert len(months) == 12
    assert min(months) == 1
    assert max(months) == 12


def test_expand_month_days_range_3_months(hash_ids):
    months = set()
    expression = 'H H * H/3 *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _months = expanded[0][3]
        assert len(_months) == 4
        months.update(_months)
    assert len(months) == 12
    assert min(months) == 1
    assert max(months) == 12


def test_expand_month_days_range_5_months(hash_ids):
    months = set()
    expression = 'H H * H/5 *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _months = expanded[0][3]
        assert len(_months) in {2, 3}
        months.update(_months)
    assert len(months) == 12
    assert min(months) == 1
    assert max(months) == 12


def test_expand_months_with_full_range(hash_ids):
    months = set()
    expression = '* * * H(1-12) *'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        months.add(expanded[0][3][0])
    assert len(months) == 12
    assert min(months) == 1
    assert max(months) == 12


# Week days expander tests
def test_expand_week_days(hash_ids):
    week_days = set()
    expression = 'H H * * H'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        week_days.add(expanded[0][4][0])
    assert len(week_days) == 7
    assert min(week_days) == 0
    assert max(week_days) == 6


def test_expand_week_days_range_2_days(hash_ids):
    days = set()
    expression = 'H H * * H/2'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _days = expanded[0][4]
        assert len(_days) in {3, 4}
        days.update(_days)
    assert len(days) == 7
    assert min(days) == 0
    assert max(days) == 6


def test_expand_week_days_range_4_days(hash_ids):
    days = set()
    expression = 'H H * * H/4'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        _days = expanded[0][4]
        assert len(_days) in {1, 2}
        days.update(_days)
    assert len(days) == 7
    assert min(days) == 0
    assert max(days) == 6


def test_expand_week_days_with_full_range(hash_ids):
    days = set()
    expression = '* * * * H(0-6)'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        days.add(expanded[0][4][0])
    assert len(days) == 7
    assert min(days) == 0
    assert max(days) == 6


# Years expander tests
def test_expand_years_by_division(hash_ids):
    years = set()
    year_min, year_max = croniter.RANGES[6]
    expression = '* * * * * * H/10'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        assert len(expanded[0][6]) == 13
        years.update(expanded[0][6])
    assert len(years) == year_max - year_min + 1
    assert min(years) == year_min
    assert max(years) == year_max


def test_expand_years_by_range(hash_ids):
    years = set()
    expression = '* * * * * * H(2020-2030)'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        years.add(expanded[0][6][0])
    assert len(years) == 11
    assert min(years) == 2020
    assert max(years) == 2030


def test_expand_years_by_range_and_division(hash_ids):
    years = set()
    expression = '* * * * * * H(2020-2050)/10'
    for hash_id in hash_ids:
        expanded = croniter.expand(expression, hash_id=hash_id)
        years.update(expanded[0][6])
    assert len(years) == 31
    assert min(years) == 2020
    assert max(years) == 2050


if __name__ == '__main__':
    pytest.main()
